#!/usr/bin/env node

'use strict';

const chalk = require('chalk');
const commander = require('commander');
const prompt = require('prompt');
const optimist = require('optimist');
const path = require('path');
const execSync = require('child_process').execSync;
const builtins = require('builtins')();
const fs = require('fs-extra');
const os = require('os');
const semver = require('semver');
const spawn = require('cross-spawn');
const _ = require('lodash');
const packageJson = require('./package.json');

const nodeMinVersion = '6.0.0';
const npmMinVersion = '6.1.0';

var schema = {
    properties: {
        name: {
            description: 'package name',
            pattern: /^[a-z0-9A-Z\-]+$/,
            required: true
        },
        description: {
            description: 'description',
            required: true
        },
        author: {
            description: 'author',
            required: true
        }
    }
};

let author = getAuthorInfo();
if (author) {
    //console.log(author);
    schema.properties.author['default'] = author;
}

const argv = optimist
    .usage(`Usage: npm init mojo-lambda-app [options]`)
    .options('h', {
        alias: 'help',
        describe: 'show help'
    })
    .options('v', {
        alias: 'version',
        describe: 'show version'
    })
    .options('verbose', {
        describe: 'show verbose messages'
    })
    .options('debug', {
        describe: 'show debug messages'
    })
    .options('name', {
        describe: 'app name to create'
    })
    .options('description', {
        describe: 'description about app to create'
    })
    .argv;

if (argv.help) {
    printLog();
    printLog(optimist.help());
} else if (argv.version) {
    printLog(packageJson.version);
} else {
    prompt.override = argv;
    prompt.start();
    prompt.get(schema, function (err, result) {
        printLog();

        if (err) {
            printLog(`Aborted!`);
        } else {

            checkNodeVersion();
            checkNpmVersion();

            const appName = result.name;
            const appDescription = result.description;
            const appAuthor = result.author;

            checkAppName(appName);

            const ownPath = path.resolve(__dirname);
            const appPath = path.resolve(appName);

            printLog(`${chalk.underline('Version')}: ${chalk.yellow(packageJson.version)}`);
            printDebug('Own Path', ownPath);
            printDebug('App Path', appPath);
            printLog();
            printLog(`Creating a new lambda app in ${chalk.green(appPath)}.`);
            printLog();

            printDebug('Create App Directory');
            createAppDirectory(appPath, appName);

            const template = 'standard';

            printDebug(`Copy the ${template} template`);
            copyTemplate(ownPath, appPath, appName, template);

            printDebug('Create a package.json');
            createPackageJson(appPath, appName, appDescription, appAuthor);

            const originalDirectory = process.cwd();
            process.chdir(appPath);

            printDebug('Install packages');
            installPackages(appPath);

            printLog();
            printLog(`Success! Created ${appName} at ${appPath}`);
        }
    });
}

function getAuthorInfo() {
    try {
        let username, useremail, proc; 
        
        proc = spawn.sync('git', ['config', 'user.name']);
        if (proc.status == 0) {
            username = proc.stdout;
        }
        proc = spawn.sync('git', ['config', 'user.email']);
        if (proc.status == 0) {
            useremail = proc.stdout;
        }

        let result = '';
        if (username && useremail) {
            result = `${username} <${useremail}>`;
        } else if (username) {
            result = `${username}`;
        } else if (useremail) {
            result = `${useremail}`;
        }
        return result.replace(/(\r\n\t|\n|\r\t)/g, '');
    } catch (e) {
        return false;
    }
}

function checkNodeVersion() {
    if (!semver.satisfies(process.version, '>=' + nodeMinVersion)) {
        printLog(
            chalk.yellow(
                `You are using Node ${process.version}.\n\n` +
                `Please update to Node ${nodeMinVersion} or higher.\n`
            )
        );
        process.exit(1);
    }
    printVerbose('Node version', process.version);
}

function checkNpmVersion() {
    let npmVersion = null;
    try {
        npmVersion = execSync('npm --version')
            .toString()
            .trim();
        if (!semver.gte(npmVersion, npmMinVersion)) {
            printLog(
                chalk.yellow(
                    `You are using npm ${npmVersion}.\n\n` +
                    `Please update to npm ${npmMinVersion} or higher.\n`
                )
            );
            process.exit(1);
        }
    } catch (err) {
        printLog(
            chalk.yellow(
                `Fail to identify npm version.\n\n` +
                `Please install npm ${npmMinVersion} or higher.\n`
            )
        );
        process.exit(1);
    }
    printVerbose('NPM version', npmVersion);
}

function checkAppName(appName) {
    let scopedPackagePattern = new RegExp('^(?:@([^/]+?)[/])?([^/]+?)$');
    let blacklist = [
        'node_modules',
        'favicon.ico'
    ];

    let errorMsg = `Could not create a project called ${chalk.red(`"${appName}"`)} because of npm naming restrictions:`;
    let errors = [];

    if (appName === null) {
        errors.push('name cannot be null');
        printErrorsAndExit(errorMsg, errors);
    }

    if (appName === undefined) {
        errors.push('name cannot be undefined');
        printErrorsAndExit(errorMsg, errors);
    }

    if (typeof appName !== 'string') {
        errors.push('name must be a string');
        printErrorsAndExit(errorMsg, errors);
    }

    if (!appName.length) {
        errors.push('name length must be greater than zero');
    }

    if (appName.match(/^\./)) {
        errors.push('name cannot start with a period');
    }

    if (appName.match(/^_/)) {
        errors.push('name cannot start with an underscore');
    }

    if (appName.trim() !== appName) {
        errors.push('name cannot contain leading or trailing spaces');
    }

    // No funny business
    blacklist.forEach(function (blacklistedName) {
        if (appName.toLowerCase() === blacklistedName) {
            errors.push(blacklistedName + ' is a blacklisted name');
        }
    });

    // core module names like http, events, util, etc
    builtins.forEach(function (builtin) {
        if (appName.toLowerCase() === builtin) {
            errors.push(builtin + ' is a core module name');
        }
    });

    // really-long-package-names-------------------------------such--length-----many---wow
    // the thisisareallyreallylongpackagenameitshouldpublishdowenowhavealimittothelengthofpackagenames-poch.
    if (appName.length > 50) {
        errors.push('name can no longer contain more than 50 characters')
    }

    // mIxeD CaSe nAMEs
    if (appName.toLowerCase() !== appName) {
        errors.push('name can no longer contain capital letters')
    }

    if (/[~'!()*]/.test(appName.split('/').slice(-1)[0])) {
        errors.push('name can no longer contain special characters ("~\'!()*")')
    }

    if (encodeURIComponent(appName) !== appName) {
        // Maybe it's a scoped package name, like @user/package
        var nameMatch = appName.match(scopedPackagePattern)
        if (nameMatch) {
            var user = nameMatch[1]
            var pkg = nameMatch[2]
            if (encodeURIComponent(user) === user && encodeURIComponent(pkg) === pkg) {
                printErrorsAndExit(errorMsg, errors);
            }
        }

        errors.push('name can only contain URL-friendly characters')
    }

    if (errors.length > 0) {
        printErrorsAndExit(errorMsg, errors);
    }
}

function createAppDirectory(appPath, appName) {
    fs.ensureDirSync(appName);
    if (!isSafeToCreateProjectIn(appPath, appName)) {
        process.exit(1);
    }
}

function copyTemplate(ownPath, appPath, appName, template) {
    // Copy the files for the user
    const templatePath = path.join(ownPath, 'templates', template);
    printDebug('templatePath', templatePath);
    if (fs.existsSync(templatePath)) {
        fs.copySync(templatePath, appPath);
    } else {
        console.error(
            `Could not locate supplied template: ${chalk.green(templatePath)}`
        );
        return;
    }

    // rename gitignore
    const templateGitignorePath = path.join(appPath, 'template_gitignore');
    const gitignorePath = path.join(appPath, '.gitignore');

    if (fs.existsSync(templateGitignorePath)) {
        fs.renameSync(templateGitignorePath, gitignorePath);
    }

    // rename serverless.yml
    const templateServerlessPath = path.join(appPath, 'template_serverless.yml');
    const serverlessPath = path.join(appPath, 'serverless.yml');

    if (fs.existsSync(templateServerlessPath)) {
        let content = fs.readFileSync(templateServerlessPath, 'ascii');
        content = content.replace('{serverless-project-name}', appName);
        fs.writeFileSync(serverlessPath, content);
        fs.unlinkSync(templateServerlessPath);
    }
}

function createPackageJson(appPath, appName, appDescription, appAuthor) {
    let packageJson = {
        name: appName,
        description: appDescription,
        author: appAuthor
    };

    const templatePackageJsonPath = path.join(appPath, 'template_package.json');

    if (fs.existsSync(templatePackageJsonPath)) {
        const templatePackageJson = require(templatePackageJsonPath);
        packageJson = _.assign(packageJson, templatePackageJson);
        fs.unlinkSync(templatePackageJsonPath);
    }

    fs.writeFileSync(
        path.join(appPath, 'package.json'),
        JSON.stringify(packageJson, null, 2) + os.EOL
    );
}

function installPackages(appPath) {
    let command = 'npm';
    //let args = ['install', '--save', argv.verbose && '--verbose'].filter(e => e);

    const templatePackageDependenciesJsonPath = path.join(appPath, 'template_package_dependencies.json');

    if (fs.existsSync(templatePackageDependenciesJsonPath)) {
        //const templatePackageDependenciesJson = require(templatePackageDependenciesJsonPath).dependencies;
        const templatePackageDependenciesJson = require(templatePackageDependenciesJsonPath);

        // install dependencies
        let args = ['install', '--save', argv.verbose && '--verbose'].filter(e => e);
        args = args.concat(
            Object.keys(templatePackageDependenciesJson.dependencies).map(key => {
                return `${key}@${templatePackageDependenciesJson.dependencies[key]}`;
            })
        );

        printDebug('args', args);

        let proc = spawn.sync(command, args, { stdio: 'inherit' });
        if (proc.status !== 0) {
            printError(`\`${command} ${args.join(' ')}\` failed`);
            process.exit(1);
        }

        // install dev dependencies
        args = ['install', '--save-dev', argv.verbose && '--verbose'].filter(e => e);
        args = args.concat(
            Object.keys(templatePackageDependenciesJson.devDependencies).map(key => {
                return `${key}@${templatePackageDependenciesJson.devDependencies[key]}`;
            })
        );

        printDebug('args', args);

        proc = spawn.sync(command, args, { stdio: 'inherit' });
        if (proc.status !== 0) {
            printError(`\`${command} ${args.join(' ')}\` failed`);
            process.exit(1);
        }

        fs.unlinkSync(templatePackageDependenciesJsonPath);
    } else {
        printError(`Could not locate supplied template pacakage dependencies: ${chalk.green(templatePackageDependenciesJsonPath)}`);
        process.exit(1);
    }

    
}

function printErrorsAndExit(msg, errors) {
    printError();
    printError(msg);
    if (typeof errors !== 'undefined') {
        errors.forEach(error => {
            printError(chalk.red(`  *  ${error}`));
        });
    }
}

function isSafeToCreateProjectIn(root, name) {
    // These files should be allowed to remain on a failed install,
    // but then silently removed during the next create.
    const errorLogFilePatterns = [
        'npm-debug.log',
        'yarn-error.log',
        'yarn-debug.log',
    ];

    const validFiles = [
        '.DS_Store',
        'Thumbs.db',
        '.git',
        '.gitignore',
        '.idea',
        'README.md',
        'LICENSE',
        '.npmignore',
        'mkdocs.yml',
        'docs',
        '.travis.yml',
        '.gitlab-ci.yml',
        '.gitattributes',
    ];
    console.log();

    const conflicts = fs
        .readdirSync(root)
        .filter(file => !validFiles.includes(file))
        // Don't treat log files from previous installation as conflicts
        .filter(
            file => !errorLogFilePatterns.some(pattern => file.indexOf(pattern) === 0)
        );

    if (conflicts.length > 0) {
        console.log(
            `The directory ${chalk.green(name)} contains files that could conflict:`
        );
        console.log();
        for (const file of conflicts) {
            console.log(`  ${file}`);
        }
        console.log();
        console.log(
            'Either try using a new directory name, or remove the files listed above.'
        );

        return false;
    }

    // Remove any remnant files from a previous installation
    const currentFiles = fs.readdirSync(path.join(root));
    currentFiles.forEach(file => {
        errorLogFilePatterns.forEach(errorLogFilePattern => {
            // This will catch `(npm-debug|yarn-error|yarn-debug).log*` files
            if (file.indexOf(errorLogFilePattern) === 0) {
                fs.removeSync(path.join(root, file));
            }
        });
    });
    return true;
}

function printDebug(subject, body) {
    if (typeof (body) == 'undefined') {
        body = '';
    } else if (typeof (body) == 'object') {
        body = JSON.stringify(body);
    }
    if (argv.debug) {
        console.log(`${chalk.underline.gray(subject)}: ${chalk.gray(body)}`);
    }
}

function printVerbose(subject, body) {
    if (typeof (body) == 'undefined') {
        body = '';
    } else if (typeof (body) == 'object') {
        body = JSON.stringify(body);
    }
    if (argv.verbose || argv.debug) {
        console.log(`${chalk.underline(subject)}: ${body}`);
    }
}

function printLog(msg) {
    if (typeof (msg) == 'undefined') {
        console.log();
    } else if (typeof (msg) == 'object') {
        console.log(JSON.stringify(msg));
    } else {
        console.log(msg);
    }
}

function printError(msg) {
    if (typeof (msg) == 'undefined') {
        console.error();
    } else if (typeof (msg) == 'object') {
        console.error(JSON.stringify(msg));
    } else {
        console.error(msg);
    }
}